/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/

// main.c
// main file

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <unistd.h>
#include <errno.h>
#define _GNU_SOURCE
#include <getopt.h>
#include "main.h"
#include "glm.h"
#include "controls.h"
#include "display.h"
#include "dlists.h"
#include "camera.h"
#include "texture.h"


int hi_res;
int debug = 0;
camera_t *cam = NULL;
state_t *state = NULL;
dlists_t *dl = NULL;
texture_list *tex = NULL;
fps_t *fps = NULL;
object_t *obj_list = NULL, *last_obj = NULL;
object_t *left_shot = NULL, *right_shot = NULL, *target = NULL;
GLUquadricObj *qobj = NULL;
GLMmodel* rocket = NULL;
explosion_t *exp_list = NULL, *last_exp = NULL;
trail_t *trail_list = NULL, *last_trail = NULL;

/* array of screen resolutions */
int res[MAX_RES][2] = {{640, 480}, {800, 600}, {1024, 768}};

/* Reshape management function */
void reshape(int w, int h)
{
	if (h == 0)
		h = 1;
	glViewport(0, 0, (GLsizei) w, (GLsizei) h);
	state->windowx = w;
	state->windowy = h;
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(60.0, (GLfloat)state->windowx/(GLfloat)state->windowy, 0.5, 100.0);
}

/* Idle management function */
void idle(void)
{
	if (state->mouse_mode == MOUSE_PERSISTENT)
	{
		if (state->xprop != 0)
			rotate_around_X(fps->rot*state->xprop, cam);
		if (state->yprop != 0)
			rotate_around_Y(fps->rot*state->yprop, cam);
	}

	check_for_movement();
	glutPostRedisplay();
}

/* Visibility function */
void visibility(int vstate)
{
	if ((vstate == GLUT_VISIBLE) && (!state->pause))
		glutIdleFunc(idle);
	else
		glutIdleFunc(NULL);
}

/* Compile display lists to be used */
void set_display_lists()
{
	set_emit_off();
	set_crosshair_material();
	set_white_list();
	set_green_list();
	set_red_list();
	set_blue_list();
	set_explosion_list();
	set_shots_material();
	set_crate();
	set_compass_list();
	set_barrel_list();
	set_subfence_list();
	set_fence_list();
	set_post_list();
	set_double_post_list();
	set_console_list();
	set_container_list();
	set_door_list();
	set_fan_list();
	set_panel_list();
	set_tunnel_list();
	set_redlight_list();
	set_walls_list();
	set_window_list();
	set_ceiling_list();
	set_floor_list();
	set_monitor_base_list();
	set_monitor_list();
	set_lights_list();
	set_platform_list();
	set_ventilation_list();
	set_decor_list();
}

/* Error callback function */
void error_callback()
{
	const GLubyte *estring;
	GLenum error_code = glGetError();

	if (error_code != GL_NO_ERROR)
	{
		estring = gluErrorString(error_code);
		fprintf(stderr, "Error: %s\n", estring);
		exit(1);
	}
}

/* General GL initialization */
void gl_initialization(void)
{
	GLfloat lmodel_ambient[] = { 0.3, 0.3, 0.3, 1.0 };
	GLfloat light1_position[] = { 0.0, 0.0, 3.0, 1.0 };
	GLfloat light1_spot_direction[] = { 0.0, 0.0, -1.0 };
	GLfloat light1_ambient[] = { 1.0, 0.0, 0.0, 1.0 };
	GLfloat light1_diffuse[] = { 1.0, 0.0, 0.0, 1.0 };
	GLfloat light1_specular[] = { 1.0, 0.0, 0.0, 1.0 };
	GLfloat fogColor[] = {0.2, 0.2, 0.2, 0.4};
	
	glClearColor(0.0, 0.0, 0.0, 0.0);
	glShadeModel(GL_SMOOTH);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	
	// Lights
	glLightModelfv(GL_LIGHT_MODEL_AMBIENT, lmodel_ambient);
	glLightModeli(GL_LIGHT_MODEL_LOCAL_VIEWER, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_TWO_SIDE, GL_TRUE);
	glLightModeli(GL_LIGHT_MODEL_COLOR_CONTROL, GL_SEPARATE_SPECULAR_COLOR);

	// Light 1, headlight
	glLightfv(GL_LIGHT1, GL_POSITION, light1_position);
	glLightf(GL_LIGHT1, GL_SPOT_CUTOFF, 15.0);
	glLightfv(GL_LIGHT1, GL_SPOT_DIRECTION, light1_spot_direction);
	glLightfv(GL_LIGHT1, GL_AMBIENT, light1_ambient);
	glLightfv(GL_LIGHT1, GL_DIFFUSE, light1_diffuse);
	glLightfv(GL_LIGHT1, GL_SPECULAR, light1_specular);
	glLightf(GL_LIGHT1, GL_CONSTANT_ATTENUATION, 2.0);
//	glLightf(GL_LIGHT1, GL_LINEAR_ATTENUATION, 0.3);
//	glLightf(GL_LIGHT1, GL_QUADRATIC_ATTENUATION, 0.5);
//	glLightf(GL_LIGHT1, GL_SPOT_EXPONENT, 20.0);

	glEnable(GL_LIGHTING);

	// quadrics
	qobj = gluNewQuadric();
	MALLOC_CHECK(qobj);
#ifndef ALPSHOOT_WIN32
	gluQuadricCallback(qobj, GLU_ERROR, &error_callback);
#else
	gluQuadricCallback(qobj, GLU_ERROR, (_GLUfuncptr)&error_callback);
#endif

	// textures
	glGenTextures(AS_LASTTEXTURE, tex->tex_names);
	init_texture("data/crate_space_warn.tga", CRATE_WARN, GL_RGB);
	init_texture("data/crate_space_rckts.tga", CRATE_RCKTS, GL_RGB);
	init_texture("data/crate_space_explo.tga", CRATE_EXPLO, GL_RGB);
	init_texture("data/crate_space_vents.tga", CRATE_VENTS, GL_RGB);
	init_texture("data/flooring_metal_2.tga", FLOOR_METAL, GL_RGB);
	init_texture("data/metal_beam3.tga", BEAM_METAL, GL_RGB);
	init_texture("data/lava_rock.tga", ROCK_WALL, GL_RGB);
	init_texture("data/tech_tile.tga", FLOOR_GRATE, GL_RGB);
	init_texture("data/compass.tga", COMPASS, GL_RGB);
	init_texture("data/barrel.tga", BARREL_TOP, GL_RGB);
	init_texture("data/radio.tga", BARREL_SIDE, GL_RGB);
	init_texture("data/fence_alpha3.tga", FENCE, GL_RGBA);
	init_texture("data/panel.tga", PANEL, GL_RGB);
	init_texture("data/metal_plate_fill.tga", METAL, GL_RGB);
	init_texture("data/console1.tga", CONSOLE, GL_RGB);
	init_texture("data/console2.tga", CONSOLE2, GL_RGB);
	init_texture("data/xboxes.tga", CONTAINER, GL_RGB);
	init_texture("data/doors_metal.tga", DOOR_METAL, GL_RGB);
	init_texture("data/bumpplat.tga", BUMP_PLAT, GL_RGB);
	init_texture("data/fan.tga", FAN, GL_RGB);
	init_texture("data/dark_metal.tga", DARK_METAL, GL_RGB);
	init_texture("data/vent.tga", VENT, GL_RGB);
	init_texture("data/metal2.tga", METAL2, GL_RGB);
	init_texture("data/bronze.tga", BRONZE, GL_RGB);
	init_texture("data/white.tga", WHITE, GL_RGB);
	init_texture("data/redlight.tga", REDLIGHT, GL_RGB);
	init_texture("data/monitor1.tga", MONITOR1, GL_RGB);
	init_texture("data/monitor2.tga", MONITOR2, GL_RGB);
	init_texture("data/monitor3.tga", MONITOR3, GL_RGB);
	init_texture("data/monitor4.tga", MONITOR4, GL_RGB);
	init_texture("data/warn.tga", WARN, GL_RGB);

	// misc
	glFrontFace(GL_CW);
	glCullFace(GL_BACK);
	glEnable(GL_CULL_FACE);

	glEnable(GL_DEPTH_TEST);
	glEnable(GL_AUTO_NORMAL);
	glEnable(GL_NORMALIZE);

	glEnable(GL_BLEND);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glHint(GL_LINE_SMOOTH_HINT, GL_NICEST);
	glHint(GL_POINT_SMOOTH_HINT, GL_NICEST);
	//glHint(GL_POLYGON_SMOOTH_HINT, GL_NICEST);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
	glEnable(GL_LINE_SMOOTH);
	glEnable(GL_POINT_SMOOTH);
	//glEnable(GL_POLYGON_SMOOTH);

	// fog
	glFogi(GL_FOG_MODE, GL_EXP2);
	glFogfv(GL_FOG_COLOR, fogColor);
	glFogf(GL_FOG_DENSITY, 0.06);
	glHint(GL_FOG_HINT, GL_NICEST);
	glFogf(GL_FOG_START, 0);
	glFogf(GL_FOG_END, 20.0);
}

/* Initializes GLUT library */
void glut_initialization(int *argc, char **argv)
{
	int menu, submenu1, submenu2, submenu3;
	
	/* pass command line args to glut */
	// init size and position before calling init to let users specify
	glutInitWindowSize(state->windowx, state->windowy);
	glutInitWindowPosition(state->window_xpos, state->window_ypos);
	glutInit(argc, argv);

	glutInitDisplayMode(GLUT_DOUBLE|GLUT_RGB|GLUT_DEPTH);

	// get effective size and position as created by the window system
	glutCreateWindow(argv[0]);
	state->window_xpos = glutGet(GLUT_WINDOW_X);
	state->window_ypos = glutGet(GLUT_WINDOW_Y);
	state->windowx = glutGet(GLUT_WINDOW_WIDTH);
	state->windowy = glutGet(GLUT_WINDOW_HEIGHT);

	glutWarpPointer(state->windowx/2, state->windowy/2);

	glutSetCursor(GLUT_CURSOR_NONE);

	glutSetKeyRepeat(GLUT_KEY_REPEAT_ON);

	glutDisplayFunc(display);
	glutMouseFunc(mouse_handler);
	glutPassiveMotionFunc(passive_handler);
	glutKeyboardFunc(keyboard_handler);
	glutSpecialFunc(special_handler);
	glutKeyboardUpFunc(release_key_handler);
	glutSpecialUpFunc(special_release_handler);
	glutReshapeFunc(reshape);
//	glutMotionFunc(motion_handler);
	glutVisibilityFunc(visibility);

	// create menu
	menu = glutCreateMenu(menuSelect);
	submenu1 = glutCreateMenu(menuSelect);
	submenu2 = glutCreateMenu(menuSelect);
	submenu3 = glutCreateMenu(menuSelect);
	glutSetMenu(menu);
	glutAddMenuEntry("Menu", 0);
	glutAddMenuEntry("", 0);
	glutAddMenuEntry("[p] - Pause ", 'p');
	glutAddMenuEntry("[m] - Cycle through mouse modes ", 'm');
	glutAddMenuEntry("[x] - Toggle dual-linked cannons ", 'x');
	glutAddMenuEntry("[j] - Toggle jetpack ", 'j');
	glutAddMenuEntry("[G] - Cycle through window sizes ", 'g');
	glutAddMenuEntry("[F] - Fullscreen mode ", 'f');
	glutAddSubMenu("Shooter", submenu1);
	glutSetMenu(submenu1);
		glutAddMenuEntry("Toggle projector lines ", 'O');
		glutAddMenuEntry("Toggle trajectories ", 'T');
	glutSetMenu(menu);
	glutAddSubMenu("Lights", submenu2);
	glutSetMenu(submenu2);
		glutAddMenuEntry("[L] - Cycle main light ", 'L');
		glutAddMenuEntry("[l] - Toggle headlight ", 'l');
	glutSetMenu(menu);
	//glutAddSubMenu("Render", submenu3);
	//glutSetMenu(submenu3);
	//	glutAddMenuEntry("[k] - Toggle fog ", 'k');
	glutSetMenu(menu);
	glutAddMenuEntry("", 0);
	glutAddMenuEntry("[ESC] - Quit ", 27);
	glutMenuStatusFunc(menuStatus);
	glutAttachMenu(GLUT_RIGHT_BUTTON);
}

GLMmodel* OBJload(const char *filename)
{
	GLMmodel *model = NULL;
	char *effective_filename = NULL;
	size_t len = 0;
	int i = 0;
	
#if (!defined ALPSHOOT_WIN32)
	const char *alt_prefix[AS_NUM_ALT_PATHS] = { "/usr/share/games/alphashooter/", "/usr/local/share/games/alphashooter/", NULL };
#else
	const char *alt_prefix[AS_NUM_ALT_PATHS] = { NULL, NULL, NULL };
#endif

	len = strlen(filename)+1;

	effective_filename = (char*)malloc(len + 40);
	if (!effective_filename)
	{
		fprintf(stderr, "Error: Unable to allocate memory for effective_filename in glmReadOBJ: %s\n", strerror(errno));
		return NULL;
	}
	memset(effective_filename, 0, len+40);

	strncat(effective_filename, filename, len);

	i = 0;
	while ((i <= AS_NUM_ALT_PATHS) && (!model))
	{
		model = glmReadOBJ(effective_filename);
		if (!model)
		{
			if (alt_prefix[i])
			{
				memset(effective_filename, 0, len+40);
				strncat(effective_filename, alt_prefix[i], strlen(alt_prefix[i]));
				strncat(effective_filename, filename, len);
				i++;
			}
			else
				// just quit with an error, there's nothing else to do
				i = AS_NUM_ALT_PATHS+1;
		}
	}
	if (i == AS_NUM_ALT_PATHS+1)
	{
		fprintf(stderr, "Unable to load model %s, paths searched: %s", filename, "data/");
		for (i = 0; i < AS_NUM_ALT_PATHS; i++)
			if (alt_prefix[i])
				fprintf(stderr, ", %s", alt_prefix[i]);
		fprintf(stderr, ".\n");
	}
	
	free(effective_filename);
	return model;
}

/* Initializes structures for external models */
void init_models()
{
	rocket = OBJload("data/newrocket.obj");
	if (!rocket)
		exit(1);
	glmUnitize(rocket);
	glmScale(rocket, 0.2);
	//glmReverseWinding(rocket);
	glmFacetNormals(rocket);
	glmVertexNormals(rocket, 90.0);
}

/* Initializes objects allocating memory for them */
void init_objects()
{
	left_shot = (object_t*)malloc(sizeof(object_t));
	MALLOC_CHECK(left_shot);
	memset(left_shot, 0, sizeof(object_t));
	right_shot = (object_t*)malloc(sizeof(object_t));
	MALLOC_CHECK(right_shot);
	memset(right_shot, 0, sizeof(object_t));
	
	state = (state_t*)malloc(sizeof(state_t));
	MALLOC_CHECK(state);
	memset(state, 0, sizeof(state_t));
	dl = (dlists_t*)malloc(sizeof(dlists_t));
	MALLOC_CHECK(dl);
	memset(dl, 0, sizeof(dlists_t));
	tex = (texture_list*)malloc(sizeof(texture_list));
	MALLOC_CHECK(tex);
	memset(tex, 0, sizeof(texture_list));
	fps = (fps_t*)malloc(sizeof(fps_t));
	MALLOC_CHECK(fps);
	memset(fps, 0, sizeof(fps_t));
	fps->fps_start = (struct timeval *)malloc(sizeof(struct timeval));
	MALLOC_CHECK(fps->fps_start);
	memset(fps->fps_start, 0, sizeof(struct timeval));
	fps->fps_end = (struct timeval *)malloc(sizeof(struct timeval));
	MALLOC_CHECK(fps->fps_end);
	memset(fps->fps_end, 0, sizeof(struct timeval));

	cam = (camera_t*)malloc(sizeof(camera_t));
	MALLOC_CHECK(cam);
	memset(cam, 0, sizeof(camera_t));

	fill_object_list();
}

/* Initializes the game states */
void init_states()
{
	state->fullscreen = 0;
	state->pause = 0;
	sprintf(state->score, "%s", "0");
	memset(state->status_string, 0, 40);
	if (debug)
		strcpy(state->status_string, "Debug mode on");
	state->curr_res = RES_1024_768;
	state->windowx = 1024;
	state->windowy = 768;
	state->window_xpos = 0;
	state->window_ypos = 0;
	state->mouse_mode = MOUSE_NONE;
	state->mouse_posx = state->windowx/2;
	state->mouse_posy = state->windowy/2;
	state->xprop = 0.0;
	state->yprop = 0.0;
	state->prev_xdiff = 0;
	state->prev_ydiff = 0;
	state->light0 = 1;
	state->light1 = 0;
	state->left_cannon_fired = 0;
	state->cannons_linked = 0;
	state->shooter = 0;
	state->lines = 0;
	state->trajectories = 0;
	state->jetpack = 0;
	if (hi_res)
		state->hi_res = 1;
	else
		state->hi_res = 0;
	state->pl_ang = 0.0;
	fps->fps = 1;
	fps->shift = STEP;
	fps->rot = ROTSTEP;

	srand((long)time(NULL));
}

/* Resets camera view */
void reset_view()
{
	cam->total_rot_x = 0.0;
	cam->total_rot_y = 0.0;
	cam->total_rot_z = 0.0;
	cam->to_shift_x = 0.0;
	cam->to_shift_y = 0.0;
	cam->to_shift_z = 0.0;
	cam->old_pos.x = 0.0;
	cam->old_pos.y = 3.0;
	cam->old_pos.z = 0.0;
	cam->radius = 1.0;

	reset_camera();
}

/* Deinitializes stuff */
void deinit()
{
	int i = 0;
	
	fprintf(stdout, "Deiniting before shutdown... ");
	
	// free fps
	if (fps != NULL)
		free(fps->fps_start);
	if (fps != NULL)
		free(fps->fps_end);
	free(fps);
	
	free(state);
	
	// free display lists
	glDeleteLists(dl->emit_off_list, 1);
	glDeleteLists(dl->crosshair_list, 1);
	glDeleteLists(dl->white_list, 1);
	glDeleteLists(dl->green_list, 1);
	glDeleteLists(dl->red_list, 1);
	glDeleteLists(dl->blue_list, 1);
	glDeleteLists(dl->shots_list, 1);
	glDeleteLists(dl->crate_list, 1);
	glDeleteLists(dl->compass_list, 1);
	glDeleteLists(dl->barrel_list, 1);
	glDeleteLists(dl->sub_fence_list, 1);
	glDeleteLists(dl->fence_list, 1);
	glDeleteLists(dl->post_list, 1);
	glDeleteLists(dl->console_list, 1);
	glDeleteLists(dl->container_list, 1);
	glDeleteLists(dl->door_list, 1);
	glDeleteLists(dl->fan_list, 1);
	glDeleteLists(dl->panel_list, 1);
	glDeleteLists(dl->tunnel_list, 1);
	glDeleteLists(dl->redlight_list, 1);
	glDeleteLists(dl->walls_list, 1);
	glDeleteLists(dl->window_list, 1);
	glDeleteLists(dl->ceiling_list, 1);
	glDeleteLists(dl->floor_list, 1);
	glDeleteLists(dl->monitor_base_list, 1);
	glDeleteLists(dl->monitor_list, 1);
	glDeleteLists(dl->lights_list, 1);
	glDeleteLists(dl->platform_list, 1);
	glDeleteLists(dl->ventilation_list, 1);
	glDeleteLists(dl->decor_list, 1);
	free(dl);

	free(left_shot);
	free(right_shot);
	free(cam);
	
	// free models and textures
	if (rocket)
		glmDelete(rocket);
	for(i = 0; i < AS_MAXTEXTURES; i++)
		if (tex->tex_names[i] != 0)
			glDeleteTextures(1, &(tex->tex_names[i]));
	if (tex->tex_data)
	{
		if (tex->tex_data->data)
			free(tex->tex_data->data);
		free(tex->tex_data);
	}
	free(tex);
	
	while (obj_list != NULL)
	{
		object_t *tmp = obj_list->next;
		delete_object(obj_list);
		obj_list = tmp;
	}

	while (exp_list != NULL)
	{
		explosion_t *tmp = exp_list->next;
		free(exp_list);
		exp_list = tmp;
	}

	while (trail_list != NULL)
	{
		trail_t *tmp = trail_list->next;
		free(trail_list);
		trail_list = tmp;
	}

	gluDeleteQuadric(qobj);
	
	fprintf(stdout, "Done\n");
}

/* Prints copyright and disclaimer banner */
void print_header()
{
	fprintf(stdout, "\nAlpha Shooter %s - a 3D OpenGL First Person Shooter\n", ALPSHOOT_VERSION);
	fprintf(stdout, "Copyright (C) 2005, 2006, 2007 - Nicola Cocchiaro\n");
	fprintf(stdout, "Released under the GNU General Public License 3\n");
	fprintf(stdout, "\n");
	if (debug)
		fprintf(stdout, "Debug mode activated\n");
	fprintf(stdout, "Starting up\n");
}

/* Prints controls */
void controls()
{
	fprintf(stdout, "\nAlpha Shooter %s - a 3D OpenGL First Person Shooter\n", ALPSHOOT_VERSION);
	fprintf(stdout, "Copyright (C) 2005, 2006, 2007 - Nicola Cocchiaro\n");
	fprintf(stdout, "Released under the GNU General Public License 3\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Keyboard controls:\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Movement:\n");
	fprintf(stdout, "a\t\tmove left\n");
	fprintf(stdout, "s\t\tmove backwards\n");
	fprintf(stdout, "d\t\tmove right\n");
	fprintf(stdout, "w\t\tmove forward\n");
	fprintf(stdout, "q\t\tturn left (rotate left in flight mode)\n");
	fprintf(stdout, "e\t\tturn right (rotate right in flight mode)\n");
	fprintf(stdout, "j\t\ttoggle jetpack (flight mode)\n");
	fprintf(stdout, "r\t\tmove up (in flight mode)\n");
	fprintf(stdout, "f\t\tmove down (in flight mode)\n");
	fprintf(stdout, "left arrow\tturn left\n");
	fprintf(stdout, "right arrow\tturn right\n");
	fprintf(stdout, "up arrow\tlook up\n");
	fprintf(stdout, "down arrow\tlook down\n");
	fprintf(stdout, "TAB\t\tauto-level\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Shooter:\n");
	fprintf(stdout, "x\t\ttoggle dual-linked cannons\n");
	fprintf(stdout, "spacebar\tshoot rockets\n");
	fprintf(stdout, "g\t\tactivate object (console, close-up)\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Renderer:\n");
	fprintf(stdout, "k\t\ttoggle fog (only in debug mode)\n");
	fprintf(stdout, "l\t\ttoggle headlight\n");
	fprintf(stdout, "L\t\tcycle lighting\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Other controls:\n");
	fprintf(stdout, "p\t\tpause game\n");
	fprintf(stdout, "G\t\tcycle available window sizes\n");
	fprintf(stdout, "F\t\tfullscreen mode\n");
	fprintf(stdout, "m\t\tcycle mouse modes\n");
	fprintf(stdout, "ESC\t\tquit\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Mouse controls:\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "right button\topen menu\n");
	fprintf(stdout, "middle button\tshoot rockets\n");
	fprintf(stdout, "\n");
}

/* Prints usage message */
void usage()
{
	fprintf(stdout, "\nAlpha Shooter %s - a 3D OpenGL First Person Shooter\n", ALPSHOOT_VERSION);
	fprintf(stdout, "Copyright (C) 2005, 2006, 2007 - Nicola Cocchiaro\n");
	fprintf(stdout, "Released under the GNU General Public License 3\n");
	fprintf(stdout, "\n");
	fprintf(stdout, "Usage: game [OPTION]...\n");
	fprintf(stdout, "Options are:\n");
	fprintf(stdout, "-d | --debug\tturn debug mode on\n");
	fprintf(stdout, "-h | --help\tprint this help message\n");
	fprintf(stdout, "-H | --hi-res\tdraw more vertices on some surfaces\n");
	fprintf(stdout, "-k | --keys\tshow a list of keys available in-game\n");
	fprintf(stdout, "-v | --version\tdisplay version number\n");
	fprintf(stdout, "\n");
}

/* Main function */
int main(int argc, char **argv)
{
	int c;
	const char *optstring = "dhHkv";
	static struct option long_options[] = {
		{"debug", no_argument, 0, 'd'},
		{"help", no_argument, 0, 'h'},
		{"hi-res", no_argument, 0, 'H'},
		{"keys", no_argument, 0, 'k'},
		{"version", no_argument, 0, 'v'},
		{0, 0, 0, 0}
	};

	while ((c = getopt_long(argc, argv, optstring, long_options, NULL)) != -1)
	{
		switch(c)
		{
			case 'd':
				debug = 1;
				break;
			case 'h':
				usage();
				exit(0);
			case 'H':
				hi_res = 1;
				break;
			case 'k':
				controls();
				exit(0);
			case 'v':
				fprintf(stdout, "Version %s\n", ALPSHOOT_VERSION);
				exit(0);
			default:
				break;
		}
	}
	
	print_header();
	init_objects();
	init_states();
	atexit(deinit);
	reset_view();
	glut_initialization(&argc, argv);
	gl_initialization();
	set_display_lists();
	init_models();

	glutMainLoop();
	
	return(0);
}
