/*------------------------------------------------------------------------------
* Alpha Shooter
* A 3D OpenGL First-Person Shooter game
* Copyright (C) 2005, 2006, 2007 Nicola Cocchiaro <ncocchiaro@users.sf.net>
* Alpha Shooter can be found at http://www.sf.net/projects/alphashooter.
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation, either version 3 or, (at your option) any later
* version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
*-------------------------------------------------------------------------------
*/

// particles.c
// minimal particles engine
// some code here is public domain code by: Gustav Taxen, nv91-gta@nada.kth.se

#include "particles.h"
#include "fps.h"

/* Adds an explosion structure to the drawable explosions list */
explosion_t* add_explosion()
{
	explosion_t* new_exp = NULL;
	
	new_exp = (explosion_t*)malloc(sizeof(explosion_t));
	if (new_exp == NULL)
	{
		perror("Unable to allocate memory for new explosion list item");
		exit(1);
	}
	// if explosion list was empty
	if (exp_list == NULL)
	{
		exp_list = new_exp;
		last_exp = exp_list;
		exp_list->next = NULL;
		exp_list->prev = NULL;
	}
	else
	{
		new_exp->prev = last_exp;
		new_exp->next = NULL;
		last_exp->next = new_exp;
		last_exp = new_exp;
	}
	memset(new_exp, 0, sizeof(explosion_t));
	return(new_exp);
}

/* Deletes an explosion from drawable explosion list */
void delete_explosion(explosion_t* o)
{
	explosion_t *p = NULL, *n = NULL;

	if (o != NULL)
	{
		p = o->prev;
		n = o->next;
		if (p != NULL)
			p->next = n;
		if (n != NULL)
			n->prev = p;
		if (last_exp == o)
			last_exp = p;
		if (exp_list == o)
			exp_list = n;
		free(o);
	}
}

/* Adds an trail structure to the drawable trails list */
trail_t* add_trail()
{
	trail_t* new_trail = NULL;
	
	new_trail = (trail_t*)malloc(sizeof(trail_t));
	if (new_trail == NULL)
	{
		perror("Unable to allocate memory for new trail list item");
		exit(1);
	}
	// if trail list was empty
	if (trail_list == NULL)
	{
		trail_list = new_trail;
		last_trail = trail_list;
		trail_list->next = NULL;
		trail_list->prev = NULL;
	}
	else
	{
		new_trail->prev = last_trail;
		new_trail->next = NULL;
		last_trail->next = new_trail;
		last_trail = new_trail;
	}
	memset(new_trail, 0, sizeof(trail_t));
	return(new_trail);
}

/* Deletes an trail from drawable trail list */
void delete_trail(trail_t* o)
{
	trail_t *p = NULL, *n = NULL;

	if (o != NULL)
	{
		p = o->prev;
		n = o->next;
		if (p != NULL)
			p->next = n;
		if (n != NULL)
			n->prev = p;
		if (last_trail == o)
			last_trail = p;
		if (trail_list == o)
			trail_list = n;
		free(o);
	}
}

/* Randomizes a speed vector */
void new_speed(float dest[3], int wantNormalize)
{
	float x;
	float y;
	float z;
	float len;

	x = (2.0 * ((GLfloat) rand()) / ((GLfloat) RAND_MAX)) - 1.0;
	y = (2.0 * ((GLfloat) rand()) / ((GLfloat) RAND_MAX)) - 1.0;
	z = (2.0 * ((GLfloat) rand()) / ((GLfloat) RAND_MAX)) - 1.0;

	// Normalizing the speed vectors gives a "fireball" effect
	if (wantNormalize)
	{
		len = sqrt(x * x + y * y + z * z);

		if (len)
		{
			x = x / len;
			y = y / len;
			z = z / len;
		}
	}

	dest[0] = x;
	dest[1] = y;
	dest[2] = z;
}

/* Create a new explosion */
void new_explosion(int wantNormalize, vector vp)
{
	int i;
	explosion_t *newex = add_explosion();

	for (i = 0; i < NUM_PARTICLES; i++)
	{
		newex->particles[i].position[0] = vp.x;
		newex->particles[i].position[1] = vp.y;
		newex->particles[i].position[2] = vp.z;

		newex->particles[i].color[0] = 1.0;
		newex->particles[i].color[1] = 0.9;
		newex->particles[i].color[2] = 0.4;
		newex->particles[i].color[3] = 1.0;

		new_speed(newex->particles[i].speed, wantNormalize);
	}

	for (i = 0; i < NUM_DEBRIS; i++)
	{
		newex->debris[i].position[0] = vp.x;
		newex->debris[i].position[1] = vp.y;
		newex->debris[i].position[2] = vp.z;

		newex->debris[i].orientation[0] = 0.0;
		newex->debris[i].orientation[1] = 0.0;
		newex->debris[i].orientation[2] = 0.0;

		newex->debris[i].color[0] = 1.0;
		newex->debris[i].color[1] = 0.2;
		newex->debris[i].color[2] = 0.1;
		newex->debris[i].color[3] = 0.9;

		newex->debris[i].scale[0] = (2.0 * 
			((GLfloat) rand()) / ((GLfloat) RAND_MAX)) - 1.0;
		newex->debris[i].scale[1] = (2.0 * 
			((GLfloat) rand()) / ((GLfloat) RAND_MAX)) - 1.0;
		newex->debris[i].scale[2] = (2.0 * 
			((GLfloat) rand()) / ((GLfloat) RAND_MAX)) - 1.0;

		new_speed(newex->debris[i].speed, wantNormalize);
		new_speed(newex->debris[i].orientationSpeed, wantNormalize);
	}

	newex->fuel = 40;
}

/* Create a new rocket trail */
void new_trail(int wantNormalize, vector vp)
{
	int i;
	trail_t *trail = add_trail();
		
	for (i = 0; i < NUM_TRAIL; i++)
	{
		trail->particles[i].position[0] = vp.x;
		trail->particles[i].position[1] = vp.y;
		trail->particles[i].position[2] = vp.z;

		trail->particles[i].color[0] = 1.0;
		trail->particles[i].color[1] = 0.7;
		trail->particles[i].color[2] = 0.0;
		trail->particles[i].color[3] = 0.7;

		new_speed(trail->particles[i].speed, wantNormalize);
	}

	trail->fuel = 8;
}
